/*
 * Preprocesses groff input in UTF-8 encoding.
 * Copyright (c) Bruno Haible 2005-07-02
 * GPL
 */

/*
 * Determines which UTF-8 characters occur in the input and emits directives
 * like the following:
 *   .char \N'8364' <GROHTML:&#8364;>
 * that are recognizable in the post-grohtml output.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* Return the length (number of units) of the first character in S, putting
   its 'ucs4_t' representation in *PUC.  */
static int
u8_mbtouc_aux (unsigned int *puc, const unsigned char *s, size_t n)
{
  unsigned char c = *s;

  if (c >= 0xc2)
    {
      if (c < 0xe0)
        {
          if (n >= 2)
            {
              if ((s[1] ^ 0x80) < 0x40)
                {
                  *puc = ((unsigned int) (c & 0x1f) << 6)
                         | (unsigned int) (s[1] ^ 0x80);
                  return 2;
                }
              /* invalid multibyte character */
            }
          else
            {
              /* incomplete multibyte character */
              *puc = 0xfffd;
              return n;
            }
        }
      else if (c < 0xf0)
        {
          if (n >= 3)
            {
              if ((s[1] ^ 0x80) < 0x40 && (s[2] ^ 0x80) < 0x40
                  && (c >= 0xe1 || s[1] >= 0xa0))
                {
                  *puc = ((unsigned int) (c & 0x0f) << 12)
                         | ((unsigned int) (s[1] ^ 0x80) << 6)
                         | (unsigned int) (s[2] ^ 0x80);
                  return 3;
                }
              /* invalid multibyte character */
            }
          else
            {
              /* incomplete multibyte character */
              *puc = 0xfffd;
              return n;
            }
        }
      else if (c < 0xf8)
        {
          if (n >= 4)
            {
              if ((s[1] ^ 0x80) < 0x40 && (s[2] ^ 0x80) < 0x40
                  && (s[3] ^ 0x80) < 0x40
                  && (c >= 0xf1 || s[1] >= 0x90)
#if 1
                  && (c < 0xf4 || (c == 0xf4 && s[1] < 0x90))
#endif
                 )
                {
                  *puc = ((unsigned int) (c & 0x07) << 18)
                         | ((unsigned int) (s[1] ^ 0x80) << 12)
                         | ((unsigned int) (s[2] ^ 0x80) << 6)
                         | (unsigned int) (s[3] ^ 0x80);
                  return 4;
                }
              /* invalid multibyte character */
            }
          else
            {
              /* incomplete multibyte character */
              *puc = 0xfffd;
              return n;
            }
        }
#if 0
      else if (c < 0xfc)
        {
          if (n >= 5)
            {
              if ((s[1] ^ 0x80) < 0x40 && (s[2] ^ 0x80) < 0x40
                  && (s[3] ^ 0x80) < 0x40 && (s[4] ^ 0x80) < 0x40
                  && (c >= 0xf9 || s[1] >= 0x88))
                {
                  *puc = ((unsigned int) (c & 0x03) << 24)
                         | ((unsigned int) (s[1] ^ 0x80) << 18)
                         | ((unsigned int) (s[2] ^ 0x80) << 12)
                         | ((unsigned int) (s[3] ^ 0x80) << 6)
                         | (unsigned int) (s[4] ^ 0x80);
                  return 5;
                }
              /* invalid multibyte character */
            }
          else
            {
              /* incomplete multibyte character */
              *puc = 0xfffd;
              return n;
            }
        }
      else if (c < 0xfe)
        {
          if (n >= 6)
            {
              if ((s[1] ^ 0x80) < 0x40 && (s[2] ^ 0x80) < 0x40
                  && (s[3] ^ 0x80) < 0x40 && (s[4] ^ 0x80) < 0x40
                  && (s[5] ^ 0x80) < 0x40
                  && (c >= 0xfd || s[1] >= 0x84))
                {
                  *puc = ((unsigned int) (c & 0x01) << 30)
                         | ((unsigned int) (s[1] ^ 0x80) << 24)
                         | ((unsigned int) (s[2] ^ 0x80) << 18)
                         | ((unsigned int) (s[3] ^ 0x80) << 12)
                         | ((unsigned int) (s[4] ^ 0x80) << 6)
                         | (unsigned int) (s[5] ^ 0x80);
                  return 6;
                }
              /* invalid multibyte character */
            }
          else
            {
              /* incomplete multibyte character */
              *puc = 0xfffd;
              return n;
            }
        }
#endif
    }
  /* invalid multibyte character */
  *puc = 0xfffd;
  return 1;
}

static
#ifdef __GNUC__
__inline
#endif
int
u8_mbtouc (unsigned int *puc, const unsigned char *s, size_t n)
{
  unsigned char c = *s;

  if (c < 0x80)
    {
      *puc = c;
      return 1;
    }
  else
    return u8_mbtouc_aux (puc, s, n);
}


int main (int argc, char *argv[])
{
  char *buf;
  size_t size;
  char chars_used[0x110000/8];

  if (argc != 1)
    exit(1);

  /* Read all input. */
  buf = NULL;
  size = 0;
  {
    const int BUFSIZE = 4096;
    int alloc = 0;
    int count;

    while (!feof(stdin)) {
      if (size + BUFSIZE > alloc) {
        alloc = alloc + alloc/2;
        if (alloc < size + BUFSIZE)
          alloc = size + BUFSIZE;
        buf = realloc(buf,alloc);
        if (buf == NULL) {
          fprintf(stderr,"out of memory\n");
          exit(1);
        }
      }
      count = fread(buf+size,1,BUFSIZE,stdin);
      if (count == 0) {
        if (ferror(stdin)) {
          perror("fread");
          exit(1);
        }
      } else {
        size += count;
      }
    }
    buf = realloc(buf,size);
    if (buf == NULL) {
      fprintf(stderr,"out of memory\n");
      exit(1);
    }
  }

  /* Determine the used UTF-8 characters. */
  memset(chars_used, '\0', sizeof(chars_used));
  {
    const char *p_end = buf+size;
    const char *p;

    for (p = buf; p < p_end; ) {
      unsigned int uc;
      int bytes = u8_mbtouc(&uc,p,p_end-p);
      if (!(bytes > 0)) abort();
      if (!(uc < 0x110000)) abort();
      if (uc >= 0x80)
        chars_used[uc/8] |= 1 << (uc % 8);
      p += bytes;
    }
  }

  /* Output .char statements for the used characters. */
  {
    unsigned int uc;

    for (uc = 0x80; uc < 0x110000; uc++)
      if ((chars_used[uc/8] >> (uc % 8)) & 1)
        fprintf(stdout,".char \\N'%d' <GROHTML:&#%d;>\n",uc,uc);
  }

  /* Output the contents with references to these characters. */
  {
    const char *p_end = buf+size;
    const char *p;

    for (p = buf; p < p_end; ) {
      unsigned int uc;
      int bytes = u8_mbtouc(&uc,p,p_end-p);
      if (!(bytes > 0)) abort();
      if (!(uc < 0x110000)) abort();
      if (uc >= 0x80)
        fprintf(stdout,"\\N'%d'",uc);
      else
        fwrite(p,1,bytes,stdout);
      p += bytes;
    }
  }

  if (ferror(stdin)) {
    perror("<stdin>");
    exit(1);
  }
  if (ferror(stdout) || fclose(stdout)) {
    perror("<stdout>");
    exit(1);
  }
  exit(0);
}
