/*
 * Postprocesses groff output to UTF-8.
 * Copyright (c) Bruno Haible 2005-07-02
 * GPL
 */

/*
 * Recognizes special sequences like < ^H U ^H 2 ^H 0 ^H A ^H C ^H >
 * and converts them to the appropriate Unicode character, in UTF-8 encoding.
 * Supports only Unicode characters from U+0000 to U+FFFF because groff's
 * \N'...' syntax supports only the BMP.
 */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>

static unsigned int hex4_value (char c1, char c2, char c3, char c4)
{
  unsigned int uc = 0;
  uc += (c1 >= 'a' ? c1 - 'a' + 10 :
         c1 >= 'A' ? c1 - 'A' + 10 :
         c1 - '0');
  uc *= 16;
  uc += (c2 >= 'a' ? c2 - 'a' + 10 :
         c2 >= 'A' ? c2 - 'A' + 10 :
         c2 - '0');
  uc *= 16;
  uc += (c3 >= 'a' ? c3 - 'a' + 10 :
         c3 >= 'A' ? c3 - 'A' + 10 :
         c3 - '0');
  uc *= 16;
  uc += (c4 >= 'a' ? c4 - 'a' + 10 :
         c4 >= 'A' ? c4 - 'A' + 10 :
         c4 - '0');
  return uc;
}

/* Write the Unicode character uc. */
static void output_utf8 (unsigned int uc)
{
  char buf[4];

  if (uc < 0x80) {
    buf[0] = uc;
    fwrite(buf,1,1,stdout);
  } else if (uc < 0x800) {
    buf[1] = 0x80 | (uc & 0x3f);
    uc = uc >> 6;
    buf[0] = uc | 0xc0;
    fwrite(buf,1,2,stdout);
  } else if (uc < 0x10000) {
    buf[2] = 0x80 | (uc & 0x3f);
    uc = uc >> 6;
    buf[1] = 0x80 | (uc & 0x3f);
    uc = uc >> 6;
    buf[0] = uc | 0xe0;
    fwrite(buf,1,3,stdout);
  } else {
    buf[3] = 0x80 | (uc & 0x3f);
    uc = uc >> 6;
    buf[2] = 0x80 | (uc & 0x3f);
    uc = uc >> 6;
    buf[1] = 0x80 | (uc & 0x3f);
    uc = uc >> 6;
    buf[0] = uc | 0xf0;
    fwrite(buf,1,4,stdout);
  }
}

int main (int argc, char *argv[])
{
  char buf[13];

  if (argc != 1)
    exit(1);

  for (;;) {
    int c;

    c = getc(stdin);
    if (c == EOF)
      break;
  restart1:
    buf[0] = c;
    if (c == '<') {
      c = getc(stdin);
      if (c == EOF) {
        fwrite(buf,1,1,stdout);
        break;
      }
      buf[1] = c;
      if (c == 0x08) {
        c = getc(stdin);
        if (c == EOF) {
          fwrite(buf,1,2,stdout);
          break;
        }
        buf[2] = c;
        if (c == 'U') {
          c = getc(stdin);
          if (c == EOF) {
            fwrite(buf,1,3,stdout);
            break;
          }
          buf[3] = c;
          if (c == 0x08) {
            c = getc(stdin);
            if (c == EOF) {
              fwrite(buf,1,4,stdout);
              break;
            }
            buf[4] = c;
            if (isxdigit(c)) {
              c = getc(stdin);
              if (c == EOF) {
                fwrite(buf,1,5,stdout);
                break;
              }
              buf[5] = c;
              if (c == 0x08) {
                c = getc(stdin);
                if (c == EOF) {
                  fwrite(buf,1,6,stdout);
                  break;
                }
                buf[6] = c;
                if (isxdigit(c)) {
                  c = getc(stdin);
                  if (c == EOF) {
                    fwrite(buf,1,7,stdout);
                    break;
                  }
                  buf[7] = c;
                  if (c == 0x08) {
                    c = getc(stdin);
                    if (c == EOF) {
                      fwrite(buf,1,8,stdout);
                      break;
                    }
                    buf[8] = c;
                    if (isxdigit(c)) {
                      c = getc(stdin);
                      if (c == EOF) {
                        fwrite(buf,1,9,stdout);
                        break;
                      }
                      buf[9] = c;
                      if (c == 0x08) {
                        c = getc(stdin);
                        if (c == EOF) {
                          fwrite(buf,1,10,stdout);
                          break;
                        }
                        buf[10] = c;
                        if (isxdigit(c)) {
                          c = getc(stdin);
                          if (c == EOF) {
                            fwrite(buf,1,11,stdout);
                            break;
                          }
                          buf[11] = c;
                          if (c == 0x08) {
                            c = getc(stdin);
                            if (c == EOF) {
                              fwrite(buf,1,12,stdout);
                              break;
                            }
                            buf[12] = c;
                            if (c == '>') {
                              /* Found one! */
                              output_utf8(hex4_value(buf[4],buf[6],buf[8],buf[10]));
                            } else {
                              fwrite(buf,1,12,stdout);
                              goto restart1;
                            }
                          } else if (c == '>') {
                            /* Found one! */
                            output_utf8(hex4_value(buf[4],buf[6],buf[8],buf[10]));
                          } else {
                            fwrite(buf,1,11,stdout);
                            goto restart1;
                          }
                        } else {
                          fwrite(buf,1,10,stdout);
                          goto restart1;
                        }
                      } else {
                        fwrite(buf,1,9,stdout);
                        goto restart1;
                      }
                    } else {
                      fwrite(buf,1,8,stdout);
                      goto restart1;
                    }
                  } else {
                    fwrite(buf,1,7,stdout);
                    goto restart1;
                  }
                } else {
                  fwrite(buf,1,6,stdout);
                  goto restart1;
                }
              } else {
                fwrite(buf,1,5,stdout);
                goto restart1;
              }
            } else {
              fwrite(buf,1,4,stdout);
              goto restart1;
            }
          } else {
            fwrite(buf,1,3,stdout);
            goto restart1;
          }
        } else {
          fwrite(buf,1,2,stdout);
          goto restart1;
        }
      } else {
        fwrite(buf,1,1,stdout);
        goto restart1;
      }
    } else {
      fwrite(buf,1,1,stdout);
    }
  }

  if (ferror(stdin)) {
    perror("<stdin>");
    exit(1);
  }
  if (ferror(stdout) || fclose(stdout)) {
    perror("<stdout>");
    exit(1);
  }
  exit(0);
}
